package portforward

import (
	"errors"
	"net/http"
	"os"

	"k8s.io/apimachinery/pkg/types"
	"k8s.io/client-go/tools/portforward"
	"k8s.io/client-go/transport/spdy"
	cmdutil "k8s.io/kubectl/pkg/cmd/util"
)

type PortForwarder struct {
	f            cmdutil.Factory
	podRef       types.NamespacedName
	port         string // local:destination
	stopChannel  chan struct{}
	readyChannel chan struct{}
	pf           *portforward.PortForwarder
}

func New(f cmdutil.Factory, podRef types.NamespacedName, port string) PortForwarder {
	return PortForwarder{
		f:      f,
		port:   port,
		podRef: podRef,
	}
}

func (pf *PortForwarder) Start() error {
	if pf.pf != nil {
		return errors.New("port already forwarding")
	}
	pf.stopChannel = make(chan struct{}, 1)
	pf.readyChannel = make(chan struct{})
	restClient, err := pf.f.RESTClient()
	if err != nil {
		return err
	}
	config, err := pf.f.ToRawKubeConfigLoader().ClientConfig()
	if err != nil {
		return err
	}
	transport, upgrader, err := spdy.RoundTripperFor(config)
	if err != nil {
		return err
	}
	req := restClient.Post().
		Resource("pods").
		Namespace(pf.podRef.Namespace).
		Name(pf.podRef.Name).
		SubResource("portforward")

	dialer := spdy.NewDialer(upgrader, &http.Client{Transport: transport}, "POST", req.URL())
	portForwarder, err := portforward.New(dialer, []string{pf.port}, pf.stopChannel, pf.readyChannel, os.Stdout, os.Stderr)
	if err != nil {
		return err
	}
	pf.pf = portForwarder
	go pf.forward()
	<-pf.readyChannel
	return nil
}

func (pf *PortForwarder) forward() {
	err := pf.pf.ForwardPorts()
	if err != nil {
		println(err)
	}
}

func (pf *PortForwarder) Close() {
	pf.stopChannel <- struct{}{}
	pf.pf = nil
}
