package rocketchat

import (
	"context"
	"errors"
	"fmt"

	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/types"
	"libre.sh/cli/pkg/portforward"
)

type MongoServer struct {
	*RocketChatInstance
	podRef    types.NamespacedName
	localPort string
	pf        *portforward.PortForwarder
}

type MongoHelloCommandResult struct {
	IsWritablePrimary bool `json:"isWritablePrimary"`
}

func (rci *RocketChatInstance) MongoSecret() (corev1.Secret, error) {
	if rci.mongoSecret != nil {
		return *rci.mongoSecret, nil
	}
	client, err := rci.KubernetesClientSet()
	if err != nil {
		return corev1.Secret{}, err
	}
	secret, err := client.CoreV1().Secrets(rci.deploymentRef.Namespace).Get(context.TODO(), fmt.Sprintf("%s-app", rci.Name), metav1.GetOptions{})
	if err != nil {
		return corev1.Secret{}, err
	}
	rci.mongoSecret = secret
	return *secret, nil
}

func (rci *RocketChatInstance) MongoAdminCredentials() (string, string, error) {
	secret, err := rci.MongoSecret()
	if err != nil {
		return "", "", err
	}
	user := "rocketchat"
	// user := secret.Data["MONGODB_USER_ADMIN_USER"]
	// decodedUser := make([]byte, len(encodedUser))
	// _, err = base64.RawStdEncoding.Decode(decodedUser, encodedUser)
	// if err != nil {
	// 	return "", "", err
	// }
	password := secret.Data["MONGO_PASSWORD"]
	// decodedPassword := make([]byte, len(encodedPassword))
	// _, err = base64.RawStdEncoding.Decode(decodedPassword, encodedPassword)
	// if err != nil {
	// 	return "", "", err
	// }
	// return string(decodedUser), string(decodedPassword), nil
	return string(user), string(password), nil
}

func (rci *RocketChatInstance) MongoServers() ([]MongoServer, error) {
	if rci.mongoServers != nil {
		return rci.mongoServers, nil
	}
	client, err := rci.KubernetesClientSet()
	if err != nil {
		return nil, err
	}
	pods, err := client.CoreV1().Pods(rci.deploymentRef.Namespace).List(context.TODO(), metav1.ListOptions{
		LabelSelector: fmt.Sprintf("app.kubernetes.io/component=mongod,app.kubernetes.io/instance=%s", rci.Name),
	})
	if err != nil {
		return nil, err
	}
	list := []MongoServer{}
	for _, v := range pods.Items {
		list = append(list, MongoServer{
			RocketChatInstance: rci,
			podRef: types.NamespacedName{
				Name:      v.Name,
				Namespace: v.Namespace,
			},
			localPort: "27017",
		})
	}
	rci.mongoServers = list
	return list, nil
}

func (rci *RocketChatInstance) MongoPrimaryServer() (MongoServer, error) {
	if rci.mongoPrimaryServer != nil {
		return *rci.mongoPrimaryServer, nil
	}
	servers, err := rci.MongoServers()
	if err != nil {
		return MongoServer{}, err
	}
	for _, server := range servers {
		client, err := server.MongoClient()
		if err != nil {
			return MongoServer{}, err
		}
		res := client.Database("rocketchat").RunCommand(context.TODO(), bson.D{{"isMaster", 1}})
		server.Close()
		var doc bson.M
		err = res.Decode(&doc)
		if err != nil {
			return MongoServer{}, err
		}
		primary := doc["ismaster"].(bool)
		if primary {
			rci.mongoPrimaryServer = &server
			return server, nil
		}
	}
	return MongoServer{}, errors.New("mongo : no primary found")
}

func (mc *MongoServer) MongoUrl() (string, error) {
	user, pass, err := mc.MongoAdminCredentials()
	if err != nil {
		return "", err
	}
	url := fmt.Sprintf("mongodb://%s:%s@localhost:%s/rocketchat?authSource=admin&replicaSet=rs0&w=majority", user, pass, mc.localPort)
	return url, nil
}

func (mc *MongoServer) MongoClient() (*mongo.Client, error) {
	url, err := mc.MongoUrl()
	if err != nil {
		return nil, err
	}
	clientOptions := options.Client().ApplyURI(url).SetDirect(true)
	if mc.pf == nil {
		pf := portforward.New(mc, mc.podRef, "27017:27017")
		err := pf.Start()
		if err != nil {
			pf.Close()
			return nil, err
		}
		mc.pf = &pf
	}
	client, e := mongo.Connect(context.TODO(), clientOptions)
	if e != nil {
		mc.Close()
		return nil, e
	}
	e = client.Ping(context.TODO(), nil)
	if e != nil {
		mc.Close()
		return nil, e
	}
	return client, nil
}

func (mc *MongoServer) Close() {
	mc.pf.Close()
	mc.pf = nil
}
